/*
 * This file is part of the e-field program written by Adam Cornett
 * This program is distrubuted under the General Public License.
 * The license is distrubed with the source in the file 'gpl.txt'
 */
package com.ajc.esim.core;

import static java.lang.Math.abs;
import static java.lang.Math.sqrt;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.ListIterator;

/**
 * Contains all methods for calculating and drawing field lines and potential
 * lines.
 */
class CPoint {

	int x, y;

	double ef;

	public CPoint(int x1, int y1, double ef1) {
		x = x1;
		y = y1;
		ef = ef1;
	}
}

public final class Calc {

	/**
	 * grain is the line length for drawing field lines, decreasing it does not
	 * improve resolution, the current value is the best one I have found.
	 */
	public static final double grain = 5.5;

	/** Coulomb's constant k */
	public static final double k = 8.99e9;

	/**
	 * Calculates the electric field and electric potential for a point.
	 * 
	 * @param x
	 *            X Location (in pixels)
	 * @param y
	 *            Y Location (in pixels)
	 * 
	 * @return Array with two values, first value "[0]" is the potential enegry
	 *         at point (x,y). The second value "[1]" is the electric field at
	 *         the point (x,y).
	 */
	public static double[] calcEF(int x, int y) {
		double PE = 0;
		double xforce = 0;
		double yforce = 0;
		double[] rv = new double[2];
		for (Charge temp : Launcher.dat.getCharges()) {
			double tX = temp.getX();
			double tY = temp.getY();
			double dist = dist(tX, tY, x, y);
			double claw = cLaw(temp.getCharge(), dist);
			yforce += (claw * ((tY - y) / dist));
			xforce += (claw * ((tX - x) / dist));
			PE += potential(temp.getCharge(), dist);
		}
		rv[0] = PE;
		rv[1] = sqrt((xforce * xforce) + (yforce * yforce));
		return rv;
	}

	/**
	 * Coulomb's law calculator
	 * 
	 * @param charge
	 *            PointCharge with which to calculate the charge
	 * @param radius
	 *            Distance from the charge(in pixels) to calculate the force at.
	 * 
	 * @return double, force in Newtons exerted by the charge
	 */
	public static double cLaw(double charge, double radius) {
		radius /= Launcher.dat.siFactor; // conversion factor
		return (k * -charge) / (radius * radius);
	}

	/**
	 * Pythagorean Therom - Calculates the distance between two points: (x1,y1)
	 * and (x2,y2).
	 * 
	 * @param x1
	 *            X Location of first element
	 * @param y1
	 *            Y Location of first element
	 * @param x2
	 *            X Location of second element
	 * @param y2
	 *            Y Location of second element
	 * 
	 * @return distance between two points
	 */
	public static double dist(double x1, double y1, double x2, double y2) {
		return sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1)));
	}

	/**
	 * Draws a curve representing the electric field line of a group of
	 * charge(s). Delegates calculations to totalForce. The method draws a line
	 * from point c (the starting point) to point n, with n being the grain the
	 * percentage of the force up/down, left/right the color of the line segment
	 * is calculated by valuGrad the loop stops if line is outside of a fixed
	 * window (-1600 to 1600 pixels).
	 * 
	 * @param x
	 *            X location to begin drawing
	 * @param y
	 *            Y location to begin drawing
	 * @param g
	 *            Graphics2D used to draw
	 * 
	 * @see #valuGrad(double, double, double)
	 * @see #totalForce2(int, int)
	 */
	public synchronized static ArrayList<CPoint> drawFieldLine(int x, int y) {
		/* current x */
		int cX = x;
		/* current y */
		int cY = y;
		/* next x */
		int nX = 0;
		/* next y */
		int nY = 0;
		/* percentage change in x and y values (between -1 and +1) */
		double xper = 0;
		/* percentage change in x and y values (between -1 and +1) */
		double yper = 0;
		ArrayList<CPoint> pts = new ArrayList<CPoint>();
		while (checkWindow(cX, cY) && !prox(cX, cY)) {
			double[] temp = totalForce2(cX, cY);
			xper = temp[1] / temp[0];
			yper = temp[2] / temp[0];
			nX = cX + (int) (grain * xper);
			nY = cY + (int) (grain * yper);
			pts.add(new CPoint(nX, nY, temp[0]));
			cX = nX;
			cY = nY;
		}
		return pts;
	}

	public synchronized static void drawLines(ArrayList<CPoint> pts,
			Graphics2D g) {
		ListIterator<CPoint> p = pts.listIterator();
		CPoint p1;
		if (p.hasNext())
			p1 = p.next();
		else
			return;
		CPoint p2;
		while (p.hasNext()) {
			p2 = p.next();
			g.setColor(valuGrad(Launcher.dat.gradMin, Launcher.dat.gradMax,
					Math.log(abs(p2.ef))));
			g.drawLine(p1.x, p1.y, p2.x, p2.y);
			p1 = p2;
		}
	}

	/**
	 * Draws a 'line'(thin graident) where the points on that line all have the
	 * same electric potential
	 * 
	 * @param val
	 *            Potential at which to draw the line
	 * @param g1
	 *            Graphics2D with which to paint
	 */
	public static void drawPELine(double val, Graphics2D g1) {
		/*
		 * TODO:Color PE Lines according to PE value
		 */
		int w = Launcher.win.getEFWidth();
		int h = Launcher.win.getEFHeight();
		int res = Launcher.dat.getPeRes();
		g1.setColor(Launcher.dat.potentialColor);
		/*
		 * These values are range values set through experimentation, there is a
		 * force max and min(fMax,fMin) and a multiplier min and max(multMin,
		 * multMax).
		 */
		final double multMin = 1e-6;
		final double multMax = 2;
		final double fMin = 50;
		final double rmult = 5e7;
		final double fMax = 5e7;
		final double multRange = (multMax - multMin);
		final double fRange = (fMax - fMin);
		double fPer = .5;
		double range = rmult;
		for (int x = 0; x < w; x += res) {
			for (int y = 0; y < h; y += res) {
				double[] var = calcEF(x, y);
				double dist = abs(var[0] - val);
				fPer = (var[1] - fMin) / fRange;
				range = rmult * fPer * multRange;
				if (dist < range) {
					g1.setColor(valuGradBW(0, range, (dist)));
					g1.fillRect(x, y, res, res);
				}
			}
		}
	}

	/**
	 * Gets the best prefix for a number.
	 * 
	 * @param val
	 *            number to look for a prefix for
	 * 
	 * @return Prefix
	 * 
	 * @see Data#getPrefix(int)
	 */
	public static Prefix findPrefix(double val) {
		int power = (int) Math.log10(abs(val));
		return Launcher.dat.getPrefix(power);
	}

	/**
	 * Determines if a value is with in a range of a target value
	 * 
	 * @param val
	 *            value to check
	 * @param target
	 *            target value
	 * @param plusminus
	 *            'error term' maximum absloute distance from the target value.
	 * 
	 * @return true if val is within the error term for the targe value, false
	 *         otherwise.
	 */
	public static boolean plus_minus(double val, double target, double plusminus) {
		boolean rv = false;
		if (((val + plusminus) < target) && ((val - plusminus) > target)) {
			return true;
		}
		return rv;
	}

	/**
	 * Potential enegry formula
	 * 
	 * @param charge
	 *            double, charge (in Coulombs)
	 * @param radius
	 *            double, distance from charge to calculate PE at
	 * 
	 * @return double, Voltage at a given distance from a charge.
	 */
	public static double potential(double charge, double radius) {
		radius /= Launcher.dat.siFactor; // conversion factor in pixels/meter
		return (k * charge) / radius;
	}

	/**
	 * Calculates the magnitude of the force (in Newtons) exerted at a point
	 * using Pythagoras' Theorem.
	 * 
	 * @param x
	 *            X Location to calculate the force at.
	 * @param y
	 *            Y Location to calculate the force at.
	 * 
	 * @return double, force at point (x,y)
	 */
	public static double totalForce(int x, int y) {
		double total = 0;
		double yf = 0;
		double xf = 0;
		for (Charge c : Launcher.dat.getCharges()) {
			double tX = c.getX();
			double tY = c.getY();
			double dist = dist(tX, tY, x, y);
			double claw = cLaw(c.getCharge(), dist);
			yf += (claw * ((tY - y) / (dist)));
			xf += (claw * ((tX - x) / (dist)));
		}
		total = sqrt((xf * xf) + (yf * yf));
		return total;
	}

	/**
	 * Returns an array containing the X force, Y Force and Total Force
	 * 
	 * @param x
	 *            X Location
	 * @param y
	 *            Y Location
	 * 
	 * @return double array, value 0 is the total force, value 1 is the x force
	 *         and value 2 is the y force
	 */
	public static double[] totalForce2(int x, int y) {
		double[] totals = new double[3];
		double yf = 0;
		double xf = 0;
		for (Charge temp : Launcher.dat.getCharges()) {
			double tX = temp.getX();
			double tY = temp.getY();
			double dist = dist(tX, tY, x, y);
			double claw = cLaw(temp.getCharge(), dist);
			if (temp.getType() == Charge.ChargeType.PointCharge) {
				yf += (claw * ((tY - y) / (dist)));
				xf += (claw * ((tX - x) / (dist)));
			}
		}
		totals[0] = sqrt((xf * xf) + (yf * yf));
		totals[1] = xf;
		totals[2] = yf;
		return totals;
	}

	/**
	 * Returns the total force and angle.
	 * 
	 * @param x
	 *            X Location
	 * @param y
	 *            Y Location
	 * 
	 * @return double array value 0 is the total force, value 1 is the angle (in
	 *         degrees)
	 */
	public static double[] totalForceAngle(int x, int y) {
		double[] total = new double[2];
		double yf = 0;
		double xf = 0;
		for (Charge c : Launcher.dat.getCharges()) {
			double tX = c.getX();
			double tY = c.getY();
			double dist = dist(tX, tY, x, y);
			double claw = cLaw(c.getCharge(), dist);
			yf += (claw * ((tY - y) / (dist)));
			xf += (claw * ((tX - x) / (dist)));
		}
		total[0] = sqrt((xf * xf) + (yf * yf));
		total[1] = Math.toDegrees(Math.atan(xf / yf));
		return total;
	}

	/**
	 * Total potential enegery at a point.
	 * 
	 * @param x
	 *            X Location
	 * @param y
	 *            Y Location
	 * 
	 * @return double, total potential enegry at point (x,y)
	 */
	public static double totalPE(int x, int y) {
		double PE = 0;
		for (Charge temp : Launcher.dat.getCharges()) {
			PE += potential(temp.getCharge(), dist(x, y, temp.getX(), temp
					.getY()));
		}
		return PE;
	}

	/**
	 * Returns a string representation of a number truncated to a specificed
	 * number of places
	 * 
	 * @param val
	 *            Number to truncate
	 * @param places
	 *            Number of places to keep
	 * 
	 * @return String
	 * 
	 * @see #truncate(double, int)
	 */
	public static String toTruncString(double val, int places) {
		String rv = "";
		Prefix p = findPrefix(val);
		String PF = p.prefix;
		double valts = (val / p.mult());
		rv = truncate(valts, places) + " " + PF;
		return rv;
	}

	/**
	 * Truncates a number to the specificed number of decimal places
	 * 
	 * @param val
	 *            Number to be truncated
	 * @param deci
	 *            Number of decimal places to be kept
	 * 
	 * @return double, truncated to deci number of decimal places
	 */
	public static double truncate(double val, int deci) {
		double rv;
		double pow = Math.pow(10, deci);
		val *= pow;
		double tempd = ((long) val) / pow;
		rv = tempd;
		return rv;
	}

	/**
	 * Calculates color based on the input value's location in the range
	 * 
	 * @param min
	 *            minimum value of range
	 * @param max
	 *            maximum value of range
	 * @param val
	 *            value to find color for
	 * 
	 * @return java.awt.Color
	 */
	public static Color valuGrad(double min, double max, double val) {
		int red = 0, green = 0, blue = 0;
		if (val >= max)
			return Color.RED;
		if (val < min)
			return Color.BLUE;
		double per = (val - min) / (max - min);
		red = (int) (255 * (per));
		blue = (int) (255 * (1 - per));
		green = (int) (255 * (1 - Math.abs(1 - (2 * per))));
		return new Color(red, green, blue);
	}

	/**
	 * Modified version of valuGrad for the drawPELine method. Returns an alpha
	 * gradient using the potential color.
	 * 
	 * @param min
	 *            Minimum value, if val==min, alpha value is 0: Color is
	 *            transparent.
	 * @param max
	 *            Maximum value, if val==max alpha value is 255: Color is fully
	 *            opaque.
	 * @param val
	 *            value, must between min and max or color returned will be
	 *            null!
	 * 
	 * @return java.awt.Color, same color as Launcher.dat.potentialColor, alpha
	 *         value equal to val as a percentage of max.
	 */
	public static Color valuGradBW(double min, double max, double val) {
		if ((val < min) || (val > max)) {
			return null;
		}
		int perval = (int) (255 * ((val - min) / (max - min)));
		int r = Launcher.dat.potentialColor.getRed();
		int g = Launcher.dat.potentialColor.getGreen();
		int b = Launcher.dat.potentialColor.getBlue();
		Color temp;
		temp = new Color(r, g, b, 255 - perval);
		return temp;
	}

	/**
	 * Calculates the angle clockwise from verticle that the force points
	 * 
	 * @param x
	 *            X location to determine the angle
	 * @param y
	 *            Y location to determine the angle
	 * 
	 * @return degrees
	 */
	public static double vectorAngle(int x, int y) {
		double theta = Math.atan(xForce(x, y) / yForce(x, y));
		return Math.toDegrees(theta);
	}

	/**
	 * Calculates the horizontal force exerted by all the charges at a given
	 * point
	 * 
	 * @param x
	 *            X Location to calculate the force at
	 * @param y
	 *            Y Location to calculate the force at
	 * 
	 * @return Force in Newtons
	 */
	public static double xForce(int x, int y) {
		double xforce = 0;
		for (Charge temp : Launcher.dat.getCharges()) {
			double tX = temp.getX();
			double tY = temp.getY();
			double dist = dist(x, y, tX, tY);
			if (temp.getType() == Charge.ChargeType.PointCharge) {
				xforce += (cLaw(temp.getCharge(), dist) * ((tX - x) / (dist)));
			}
		}
		return xforce;
	}

	/**
	 * Calculates the verticle force exerted by all the charges at a given point
	 * 
	 * @param x
	 *            X Location to calculate the force at
	 * @param y
	 *            Y Location to calculate the force at
	 * 
	 * @return Force in Newtons
	 */
	public static double yForce(int x, int y) {
		double yforce = 0;
		for (Charge temp : Launcher.dat.getCharges()) {
			double tX = temp.getX();
			double tY = temp.getY();
			double dist = dist(x, y, tX, tY);
			if (temp.getType() == Charge.ChargeType.PointCharge) {
				yforce += (cLaw(temp.getCharge(), dist) * ((tX - x) / (dist)));
			}
		}
		return yforce;
	}

	private static boolean checkWindow(int x, int y) {
		final int boundTolerance = 800;
		if (x < -boundTolerance)
			return false;
		if ((x - Launcher.win.getWidth()) > 800)
			return false;
		if (y < -boundTolerance)
			return false;
		if ((y - Launcher.win.getHeight()) > 800)
			return false;
		return true;
	}

	/**
	 * Determines if a point is too close to a charge. Used to stop drawLine if
	 * the field line gets to another point.
	 * 
	 * @param x
	 *            X Location to check
	 * @param y
	 *            Y Location to check
	 * 
	 * @return true if point is within the radius of the charge, false otherwise
	 */
	private static boolean prox(int x, int y) {
		for (Charge temp : Launcher.dat.getCharges()) {
			if (temp.isInside(x, y) && temp.getCharge() != 0) {
				return true;
			}
		}
		return false;
	}
}
