/*
 * This file is part of the e-field program written by Adam Cornett
 * This program is distrubuted under the General Public License.
 * The license is distrubed with the source in the file 'gpl.txt'
 */
package com.ajc.esim.core;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.io.Serializable;

/**
 * Defines a basic charge, stores location, shape and charge.
 * 
 * @see sci.core.PointCharge
 * @see sci.core.LineCharge
 */
public abstract class Charge implements Comparable, Serializable {

	/**
	 * ChargeType of Charge. Should be set by implementing classes.
	 */
	public static enum ChargeType {
		LineCharge, PointCharge;
	}

	/** true if this is selected in the MainWindow window */
	public boolean isSel = false;

	/** Charge in Coulombs */
	private double charge;

	/** ChargeType of this charge, should be set by implementing classes */
	private ChargeType type = null;

	/** X Location in Pixels */
	private double x;

	/** Y Location in Pixels */
	private double y;

	/**
	 * Shape for drawing, used by paint, must be initialized by implementing
	 * sub-classes
	 */
	protected Shape myShape;

	/**
	 * Default Constructor
	 */
	public Charge() {
	}

	/**
	 * Location and Charge constructor
	 * 
	 * @param x
	 *            X Location (in pixels)
	 * @param y
	 *            Y Location (in pixels)
	 * @param c
	 *            charger (in columbs)
	 */
	public Charge(double x, double y, double c) {
		this.x = x;
		this.y = y;
		this.charge = c;
	}

	/**
	 * Checks to see if this was clicked on.
	 * 
	 * @param e
	 *            MouseEvent, passed from Launcher.win
	 * 
	 * @return true if moused clicked on this, otherwise, false
	 */
	public boolean checkClick(java.awt.event.MouseEvent e) {
		double size = getSize();
		double xLoc = e.getX();
		double yLoc = e.getY();
		boolean X = ((this.getPaintX() > (xLoc - size)) && (this.getPaintX() < (xLoc + size)));
		boolean Y = ((this.getPaintY() > (yLoc - size)) && (this.getPaintY() < (yLoc + size)));
		return (X && Y);
	}

	/**
	 * Compares this to another charge by comparing charge values
	 * 
	 * @param obj
	 *            Object, must be type Charge
	 * 
	 * @return difference in charges
	 */
	public int compareTo(Object obj) {
		int temp = 0;
		temp = (int) (this.getCharge() - ((Charge) obj).getCharge());
		return temp;
	}

	/**
	 * Draws the shape. For use while dragging the charge
	 * 
	 * @param g
	 *            Graphics for painting.
	 */
	public void drawOutline(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;
		this.paint(g2);
	}

	/**
	 * Returns the charge of this Charge object.
	 * 
	 * @return Returns the electric charge value.
	 */
	public double getCharge() {
		return charge;
	}

	/**
	 * Gets the charge in text form.
	 * 
	 * @return String formatted for display on screen.
	 */
	public String getChargeText() {
		String rv = "";
		rv += (this.charge / Launcher.dat.getScale());
		rv += (Launcher.dat.getPrefix(0).prefix + "C");
		return rv;
	}

	/**
	 * Gets the height of the charge, default is 1/2 of getSize(), can be
	 * overridden
	 * 
	 * @return double height
	 * 
	 * @see #getSize()
	 */
	public double getHeight() {
		return getSize() / 2;
	}

	/**
	 * Gets the length of the charge, default is 1/2 of getSize(), can be
	 * overridden
	 * 
	 * @return double length
	 * 
	 * @see #getSize()
	 */
	public double getLength() {
		return getSize() / 2;
	}

	/**
	 * Gets the X Location as used for painting
	 * 
	 * @return int, x location in pixels
	 */
	public int getPaintX() {
		/*
		 * TODO:Finish implementation of scaleable drawing
		 */
		return (int) (this.getX());
	}

	/**
	 * Gets the Y Location as used for painting
	 * 
	 * @return int, y location in pixels
	 */
	public int getPaintY() {
		/*
		 * TODO:Finish implementation of scaleable drawing
		 */
		return (int) (this.getY());
	}

	/**
	 * Size of charge for painting, currently set at 16.
	 * 
	 * @return double, size of charge
	 */
	public double getSize() {
		return 20;
	}

	/**
	 * 
	 * @return Returns the type of this charge.
	 */
	public ChargeType getType() {
		return type;
	}

	/**
	 * 
	 * @return Returns the x location.
	 */
	public double getX() {
		return this.x;
	}

	/**
	 * 
	 * @return Returns the y location.
	 */
	public double getY() {
		return this.y;
	}

	/**
	 * Needs to be overridden by implementing classes
	 * 
	 * @param x
	 *            X location of point to check
	 * @param y
	 *            Y location of point to check
	 * 
	 * @return boolean True if point (X,Y) is covered by the charge when
	 *         displayed on screen
	 */
	public abstract boolean isInside(double x, double y);

	/**
	 * Paint method, to be overridden by implementing classes
	 * 
	 * @param g
	 *            Graphics object for painting
	 */
	public abstract void paint(Graphics g);

	/**
	 * Builder for shape object, to be overridden by implementing classes
	 * 
	 * @see #myShape
	 */
	public abstract void reBuildShape();

	/**
	 * 
	 * @param charge
	 *            sets charge value
	 */
	public void setCharge(double charge) {
		this.charge = (charge);
	}

	/**
	 * 
	 * @param X
	 *            sets the x location value
	 *            min val of <em>-1600</em>, 
	 *            max val of <em>1600</em>
	 */
	public void setX(double X) {
		X = Math.max(-1600.0, X);
		X = Math.min(X, 1600.0);
		this.x = X;
		this.reBuildShape();
	}

	/**
	 * 
	 * @param Y
	 *            sets the y location value and rebuilds the shape
	 *            min val of <em>-1600</em>, 
	 *            max val of <em>1600</em>
	 */
	public void setY(double Y) {
		Y = Math.max(-1600.0, Y);
		Y = Math.min(Y, 1600.0);
		this.y = Y;
		this.reBuildShape();
	}

	/**
	 * A shorter version of toString();
	 * 
	 * @return String
	 */
	public String toStringShort() {
		return "(" + getChargeText() + ") at ("
				+ (int) (x / Launcher.dat.siFactor) + ","
				+ (int) (y / Launcher.dat.siFactor) + ")";
	}

	/**
	 * 
	 * @param type
	 *            The type to set.
	 */
	protected void setType(ChargeType type) {
		this.type = type;
	}
}
