/*
 * This file is part of the e-field program written by Adam Cornett
 * This program is distrubuted under the General Public License.
 * The license is distrubed with the source in the file 'gpl.txt'
 */
package com.ajc.esim.core;

import java.awt.Color;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.ListIterator;

/**
 * Stores all information for the program, written to file for program
 * persistance.
 */
public class Data implements Serializable {

	/**
	 * Array of Prefixes used for a feature still in devlopment.
	 * 
	 * @see #setupPrefixArray()
	 * @see Prefix
	 */
	public static Prefix[] prefixes = new Prefix[9];

	public Color backgroundColor = null;

	public Color gridColor = null;

	public Color infoBack = null;

	public Color negColor = null;

	public Color posColor = null;

	public Color potentialColor = null;

	public Color text = null;

	public Color tickColor = null;

	public boolean fieldLinesOn = true;

	public boolean gridOn = true;

	public boolean potentialLinesOn = false;

	public boolean ticksOn = true;

	public double siFactor = 1; /* Conversion factor for pixels/meters */

	public double totalPositiveCharge = 0;

	/*
	 * There are a whole lot of constants and variables here, most of the names
	 * are self explanatory, however if you don't know what it does, don't mess
	 * with it.
	 */
	public int MainWinHeight = 650;

	public int chargeScale = 2;

	public int mainWinWidth = 900;

	public int pLines = 5;

	public int peRes = 2;

	public int prefixIndex = 4;

	public int tickSpacing = 50;

	public double gradMax = 20;

	public double gradMin = 0;

	public long potOffset = 200000;

	public long potspacing = 350000;

	/**
	 * ArrayList for storing charges
	 * 
	 * @see #getCharge(int)
	 * @see #addCharge(Charge)
	 * @see #removeCharge(Charge)
	 * @see #removeCharge(int)
	 * @see #flushCharges()
	 */
	private volatile ArrayList<Charge> Charges = new ArrayList<Charge>();

	private int numLines = 16;

	/**
	 * Creates new Data object.
	 */
	public Data() {
		setupColors();
		setupPrefixArray();
	}

	/**
	 * Adds a charge to the Charges ArrayList
	 * 
	 * @param c
	 *            Charge to add to list
	 * 
	 * @see #Charges
	 */
	public void addCharge(Charge c) {
		Charges.add(c);
	}

	/**
	 * Clears all charges from the Charges ArrayList and then calls the System
	 * garbage collector
	 * 
	 * @see #Charges
	 */
	public void flushCharges() {
		Charges.clear();
		System.gc();
	}

	/**
	 * Gets the charge from the Charges array list at the given index.
	 * 
	 * @param index
	 *            Charges ArrayList item to return
	 * 
	 * @return Charge
	 * 
	 * @see Charge
	 * @see #Charges
	 */
	public Charge getCharge(int index) {
		return Charges.get(index);
	}

	/**
	 * Gets the index of Charge c.
	 * 
	 * @param c
	 *            Charge to find the index of
	 * 
	 * @return Index of the given Charge in the Charges ArrayList
	 * 
	 * @see #Charges
	 */
	public int getChargeIndex(Charge c) {
		return Charges.indexOf(c);
	}

	/**
	 * Returns the length of the Charges ArrayList
	 * 
	 * @return Number of Charges in the Charges ArrayList
	 * 
	 * @see #Charges
	 */
	public int getChargeLength() {
		return Charges.size();
	}

	public synchronized ArrayList<Charge> getCharges() {
		return Charges;
	}

	/**
	 * Gets the Charge density
	 * 
	 * @return Charge density (double)
	 * 
	 * @see #numLines
	 */
	public int getNumLines() {
		return numLines;
	}

	/**
	 * Gets the potential line resolution
	 * 
	 * @return Potential resolution (integer)
	 * 
	 * @see #peRes
	 */
	public int getPeRes() {
		return peRes;
	}

	/**
	 * Gets the prefix corrsponding to the specificed power of ten Currently
	 * returns only the 4th prefix (0)
	 * 
	 * @param POT
	 *            Power Of Ten
	 * 
	 * @return Prefix
	 * 
	 * @see Prefix
	 */
	public Prefix getPrefix(int POT) {
		/*
		 * currently hacked to only return the 'no prefix' prefix
		 */
		return prefixes[4];
	}

	/**
	 * Gets the index of the current Prefix
	 * 
	 * @return Index of current Prefix
	 * 
	 * @see #prefixes
	 */
	public int getPrefixIndex() {
		return prefixIndex;
	}

	/**
	 * Gets the Scale of the current prefix
	 * 
	 * @return Scaler of current prefix.
	 * 
	 * @see Prefix#mult()
	 */
	public double getScale() {
		return prefixes[prefixIndex].mult();
	}

	/**
	 * Inverts the inverted color.
	 * 
	 * @param in
	 *            Inital color
	 * 
	 * @return Inverse of inital color
	 */
	public Color negate(Color in) {
		Color newColor = null;
		newColor = new Color(255 - in.getRed(), 255 - in.getGreen(), 255 - in
				.getGreen(), 50);

		return newColor;
	}

	/**
	 * Calls reBuildShape on all charges.
	 * 
	 * @see Charge#reBuildShape()
	 */
	public void rebuildAllShapes() {
		for(Charge temp:Launcher.dat.getCharges()) {
			temp.reBuildShape();
		}
	}

	/**
	 * Removes a charge from the Charges ArrayList
	 * 
	 * @param c
	 *            Charge to be removed
	 * 
	 * @see #Charges
	 */
	public void removeCharge(Charge c) {
		Charges.remove(c);
	}

	/**
	 * Removes a charge from the Charges ArrayList based on the index number
	 * 
	 * @param index
	 *            Index of charge to be removed
	 * 
	 * @see #Charges
	 */
	public void removeCharge(int index) {
		Charges.remove(index);
	}

	/**
	 * Resets program variables to original values
	 * 
	 * @see #setupColors()
	 * @see #updateScales()
	 */
	public void resetAll() {
		this.setupColors();
		siFactor = 1;
		peRes = 2;
		pLines = 5;
		tickSpacing = 50;
		ticksOn = true;
		prefixIndex = 4;
		gridOn = true;
		updateScales();
	}

	/**
	 * Sets the Charge Density
	 * 
	 * @param numLines
	 *            New charge density
	 * 
	 * @see #numLines
	 */
	public void setNumLines(int numLines) {
		this.numLines = numLines;
	}

	/**
	 * Sets the Potential Resolution
	 * 
	 * @param peRes
	 *            New potential resolution
	 * 
	 * @see #peRes
	 */
	public void setPeRes(int peRes) {
		this.peRes = peRes;
	}

	/**
	 * Sets the current prefix index.
	 * 
	 * @param x
	 *            new prefix index; must be between 8 and 0 inclusive.
	 */
	public void setPrefix(int x) {
		x = Math.min(8, x);
		x = Math.max(0, x);
		prefixIndex = x;
	}

	/**
	 * Sets all colors to defaults.
	 */
	public void setupColors() {
		tickColor = Color.WHITE;
		potentialColor = new Color(200, 255, 200);
		text = new Color(250, 250, 255);
		backgroundColor = Color.BLACK;
		negColor = Color.BLUE;
		posColor = Color.RED;
		gridColor = negate(backgroundColor);
		infoBack = new Color(255, 255, 255, 150);
	}

	/**
	 * initalizes the prefix array
	 * 
	 * @see Prefix
	 */
	public void setupPrefixArray() {
		prefixes[8] = new Prefix("tera", "T", 12);
		prefixes[7] = new Prefix("giga", "G", 9);
		prefixes[6] = new Prefix("mega", "M", 6);
		prefixes[5] = new Prefix("kilo", "k", 3);
		prefixes[4] = new Prefix("none", "", 0);
		prefixes[3] = new Prefix("centi", "c", -2);
		prefixes[2] = new Prefix("milli", "m", -3);
		prefixes[1] = new Prefix("micro", "u", -6);
		prefixes[0] = new Prefix("nano", "n", -9);
	}

	/**
	 * Updates the potential offset and spacing, field line color scale and
	 * field line distrobution.
	 */
	public void updateScales() {
		double maxP = Double.MIN_VALUE;
		double maxF = Double.MIN_VALUE;
		double posTotal = 0;
		for(Charge temp:Launcher.dat.getCharges()) {
			double tcharge = temp.getCharge();
			if (tcharge > 0)
				posTotal += tcharge;
			double[] t = Calc.calcEF((int) temp.getX() + 6,
					(int) temp.getY() + 6);
			maxP = Math.max(maxP, t[0]);
			maxF = Math.max(maxF, t[1]);
		}
		if (maxP > 0) {
			this.potOffset = (long) maxP / 8;
			this.potspacing = potOffset / pLines;
		}
		if (maxF > 0) {
			this.gradMax = (long) Math.log(maxF * 2);
			this.gradMin = 0;
		}

		this.totalPositiveCharge = posTotal;
	}

	public Charge[] getChargeArray() {
		Charge[] rv = new Charge[Charges.size()];
		Charges.toArray(rv);
		return rv;
	}
}
