/*
 * This file is part of the e-field program written by Adam Cornett
 * This program is distrubuted under the General Public License.
 * The license is distrubed with the source in the file 'gpl.txt'
 */
package com.ajc.esim.core;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.logging.Level;
import java.util.logging.SimpleFormatter;

import javax.swing.JDialog;
import javax.swing.JFrame;

import com.ajc.esim.core.logging.AFileHandler;
import com.ajc.esim.core.logging.ILogger;
import com.ajc.esim.core.logging.Lev;
import com.ajc.esim.core.logging.SLogger;
import com.ajc.esim.gui.MainWindow;
import com.ajc.esim.gui.Messages;
import com.ajc.esim.gui.Splash;

/**
 * Entry point for the program,contains the main window, a data class and starts
 * a background thread for repainting the screen. As of version 1.16 this class
 * also handles file saving, loading and logging. Renamed from Runtime to
 * Launcher in verison 1.27 to prevent confusion with java.lang.Runtime.
 */
public class Launcher {

	/**
	 * File Handler for logging
	 * 
	 * @see AFileHandler
	 */
	public static AFileHandler aH = null;

	/**
	 * Critical message log
	 * 
	 * @see SLogger
	 */
	public static SLogger critical = null;

	/**
	 * Data class instance, holds all application data, written to disk for
	 * application persistance.
	 * 
	 * @see Data
	 */
	public static Data dat = null;

	/**
	 * Information level log
	 * 
	 * @see ILogger
	 */
	public static ILogger info = null;

	/**
	 * If program is run as applet (usually from a web browser) then isApplet is
	 * set to true.
	 */
	public static boolean isApplet = false;

	/** On/Off for logging system */
	public static boolean LogOn = true;

	/** Program Version number, stored in external resource file. */
	public static final String version = Messages
			.getString("About.MajorVersion")
			+ "." + Messages.getString("About.MinorVersion");

	/**
	 * Instance of MainWindow window
	 * 
	 * @see MainWindow
	 */
	public static MainWindow win = null;

	/**
	 * Shutsdown the program, saving all data in dat to file 'state.o'
	 * 
	 * @param State
	 *            Shutdown state, zero(0) for normal, non zero for abnormal
	 *            shutdown
	 * 
	 * @since 1.16
	 */
	public static void Exit(int State) {
		if (!isApplet) {
			if (State == 0) {
				Save("state.o");
				if (LogOn)
					info.log(Lev.INFO, "Standard Shutdown");
			} else {
				if (LogOn)
					critical.log(Lev.SEVERE, "NON STANDARD SHUTDOWN STATE: "
							+ State);
			}
			aH.flush();
			aH.close();
			System.runFinalization();
			System.exit(State);
		} else {
			win.dispose();
			dat = null;
		}
	}

	/**
	 * Loads a Data object from file
	 * 
	 * @param file
	 *            File from which data should be read, file should be written in
	 *            manner specified in Launcher.Save. The data class written to
	 *            file must be the same version as the version currently in use.
	 *            Thus, there is not comaptability between program versions.
	 * 
	 * @return Data type object if file is valid. If an error is encountered,
	 *         null is returned and the reason for failure is writtent to the
	 *         log file.
	 * 
	 * @see #Save(String)
	 * @since 1.16
	 */
	public static Data Load(String file) {
		if (isApplet) {
			return null;
		}

		ObjectInputStream ois = null;
		FileInputStream fis = null;
		Object temp = null;

		try {
			fis = new FileInputStream(file);
		} catch (FileNotFoundException e) {
			if (LogOn) {
				info.log(Lev.FINE, "Failed to create FileInputStream", e);
			}

			return null;
		}

		try {
			ois = new ObjectInputStream(fis);
		} catch (IOException e1) {
			if (LogOn) {
				info.log(Lev.FINE, "Failed to create ObjectInputStream", e1);
			}

			return null;
		}

		try {
			temp = ois.readObject();
		} catch (IOException e2) {
			if (LogOn) {
				info.log(Lev.WARNING, "Object Read failed", e2);
			}

			return null;
		} catch (ClassNotFoundException e2) {
			if (LogOn) {
				info.log(Lev.FINE, "Failed to create Data object from File"
						+ ", invalid file or class version.", e2);
			}

			return null;
		}

		if (LogOn) {
			info.log(Lev.INFO, "Data loaded from file", dat);
		}

		Data rv = (Data) temp;

		return rv;
	}

	/**
	 * Main entry point for the program. Possible args are "NOLOG" to disable
	 * logging and "APPLET" to indicated the the progam is running as an applet.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		for (String i : args) {
			if (i.equals("NOLOG")) {
				LogOn = false;
			} else if (i.equals("APPLET")) {
				isApplet = true;
			}
		}

		/*
		 * Create Instance of Splash Screen
		 */
		Splash spl = null;
		if (!isApplet) {
			spl = new Splash();
			spl.setVisible(true);
		}

		if (LogOn) {
			info = new ILogger("Info Log", null);
			critical = new SLogger("Critical Log", null);

			try {
				/*
				 * Create log file with 1MB split point, max of 10 files.
				 */
				aH = new AFileHandler("Log", 10485760, 10, true);
			} catch (SecurityException e) {
				e.printStackTrace();
				Exit(2);
			} catch (IOException e) {
				e.printStackTrace();
				Exit(2);
			}

			/*
			 * Setup all logs
			 */
			aH.setFormatter(new SimpleFormatter());
			info.addHandler(aH);
			critical.addHandler(aH);
			info.log(Level.INFO, "E-Field Starting...");
		}

		/*
		 * Setup Program data, load from file if able, else, create new Data
		 * object.
		 */
		Data temp = null;

		try {
			temp = Load("state.o");
		} catch (java.security.AccessControlException e) {
			/*
			 * ignore if not allowed to read file.
			 */
		}

		if (temp != null) {
			dat = temp;
		} else {
			dat = new Data();
		}

		dat.setupPrefixArray();

		/*
		 * Setup MainWindow window
		 */
		JFrame.setDefaultLookAndFeelDecorated(true);
		JDialog.setDefaultLookAndFeelDecorated(true);
		try {
			win = new MainWindow();
		} catch (RuntimeException e1) {
			if (LogOn) {
				critical.log(Lev.SEVERE, "Failed to create MainWindow Window",
						e1);
			}

			Exit(1);
		}
		dat.rebuildAllShapes();
		win.setVisible(true);
		if (spl != null) {
			spl.setVisible(false);
			spl.dispose();
		}
		//test data generator
		//testSetup(10);
	}

	/**
	 * Saves the dat object to file.
	 * 
	 * @param file
	 *            File to write to
	 * 
	 * @see #Load(String)
	 * @since 1.16
	 */
	public static void Save(String file) {
		ObjectOutputStream oos = null;
		FileOutputStream fos = null;

		try {
			fos = new FileOutputStream(file);
		} catch (FileNotFoundException e) {
			if (LogOn) {
				info.log(Lev.FINE, "Failed to create FileOutputStream", e);
			}

			return;
		}

		try {
			oos = new ObjectOutputStream(fos);
		} catch (IOException e1) {
			if (LogOn) {
				info.log(Lev.FINE, "Failed to create ObjectOutputStream", e1);
			}

			return;
		}

		try {
			oos.writeObject(dat);
		} catch (IOException e2) {
			if (LogOn) {
				info.log(Lev.WARNING, "IO error while saving data", e2);
			}

			return;
		}

		if (LogOn) {
			info.log(Lev.FINE, "Data written to file");
		}
	}

	private static void testSetup(int num) {
		for (int i = 0; i < num; i++) {
			Charge c = new PointCharge(.05, 500, 200 + 4 * i);
			dat.addCharge(c);
			c = new PointCharge(-.05, 600, 200 + 4 * i);
			dat.addCharge(c);
		}
	}

}