/*
 * This file is part of the e-field program written by Adam Cornett
 * This program is distrubuted under the General Public License.
 * The license is distrubed with the source in the file 'gpl.txt'
 */
package com.ajc.esim.core;

import java.awt.Graphics;
import java.awt.Graphics2D;

import com.ajc.esim.gui.PointChargeShape;

/**
 * Defines an point charge. A point charge represents a point in space (no size)
 * that has a given electrical charge. This class holds location and charge
 * information, it also can paint itself.
 */
public class PointCharge extends Charge {

	/**
	 * Constructor, double precision
	 * 
	 * @param charg
	 *            charge in Coulombs
	 * @param X
	 *            Horiz. Location of charge
	 * @param Y
	 *            Vert. Location of charge
	 */
	public PointCharge(double charg, double X, double Y) {
		super(X, Y, charg);
		this.setType(Charge.ChargeType.PointCharge);
		this.reBuildShape();
	}

	/**
	 * Constructor, integer percision
	 * 
	 * @param charg
	 *            charge in Coulombs
	 * @param X
	 *            Horiz. Location of charge
	 * @param Y
	 *            Vert. Location of charge
	 */
	public PointCharge(double charg, int X, int Y) {
		this.setType(Charge.ChargeType.PointCharge);
		this.setCharge(charg);
		this.setX(X);
		this.setY(Y);
		this.reBuildShape();
	}

	/**
	 * Checks to seen if a point is inside of the charge as it is drawn on the
	 * screen
	 * 
	 * @param x
	 *            X Location of point to check.
	 * @param y
	 *            Y Location of point to check.
	 * 
	 * @return boolean True if point (x,y) is covered by this when it is
	 *         painted, otherwise false.
	 */
	@Override
	public boolean isInside(double x, double y) {
		if (Calc.dist(x, y, this.getX(), this.getY()) < (this.getSize() / 2)) {
			return true;
		}

		return false;
	}

	/**
	 * Draws this on screen with Graphics g (cast to a Graphics 2D object
	 * internally). Sets color based on charge sign. Draws this.myShape.
	 * 
	 * @param g
	 *            Graphics object used to paint.
	 * 
	 * @see Charge#myShape
	 */
	@Override
	public void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;

		if (this.getCharge() > 0) {
			g2.setColor(Launcher.dat.posColor);
		} else {
			g2.setColor(Launcher.dat.negColor);
		}

		g2.fill(this.myShape);
		if (isSel) {
			g2.setColor(java.awt.Color.WHITE);
			g2.draw(this.myShape);
		}

		g2.setColor(Launcher.dat.text);
		g2.drawString("" + this.getCharge(), this.getPaintX() - 9, this
				.getPaintY() + 4);
	}

	/**
	 * Creates Shape myShape(Defined in Charge). The shape is defined as a Shape
	 * in Charge, it is set as a PointChargeShape in this method.
	 * 
	 * @see sci.gui.PointChargeShape
	 * @see Charge#myShape
	 */
	@Override
	public void reBuildShape() {
		double s = getSize();
		double offSet = (s / 2);
		double dx = this.getPaintX() - offSet;
		double dy = this.getPaintY() - offSet;
		this.myShape = new PointChargeShape(dx, dy, this.getSize());
	}

	/**
	 * 
	 * @return String representing the data stored in this charge
	 */
	@Override
	public String toString() {
		return "Point Charge: " + getChargeText() + " (" + (getPaintX()) + ","
				+ (getPaintY()) + ")";
	}
}
