/*
 * This file is part of the e-field program written by Adam Cornett
 * This program is distrubuted under the General Public License.
 * The license is distrubed with the source in the file 'gpl.txt'
 */
package com.ajc.esim.gui;

import static java.lang.StrictMath.cos;
import static java.lang.StrictMath.sin;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.util.ConcurrentModificationException;
import java.util.ListIterator;

import javax.swing.JPanel;

import com.ajc.esim.core.Calc;
import com.ajc.esim.core.Charge;
import com.ajc.esim.core.Launcher;
import com.ajc.esim.core.ReDrawThread;
import com.ajc.esim.core.logging.Lev;

public class EFPanel extends JPanel {

	public volatile boolean reDrawBackground = true;

	public volatile boolean restartPaint = false;

	private Image Background;

	private volatile Graphics2D o = null;

	private Image offScreen;

	private Image offScreen1;

	private ReDrawThread rdt = new ReDrawThread();

	private int res = Launcher.dat.getPeRes();

	/**
	 * Creates a new EFPanel object.
	 */
	public EFPanel() {
		super();
		this.setDoubleBuffered(true);
		this.addKeyListener(Launcher.win);
		initialize();
	}

	public void drawOffScreen() {
		Launcher.win.setStatusLabel("Please wait, redrawing...");
		if (rdt.isAlive())
			rdt.interrupt();
		while (rdt.isAlive())
			;
		if (!rdt.isAlive()) {
			rdt = new ReDrawThread();
			rdt.start();
		}
	}

	public synchronized void finishDraw() {
		clearScreen((Graphics2D) offScreen1.getGraphics());
		offScreen1.getGraphics().drawImage(offScreen, 0, 0, this);
	}

	public int getRes() {
		return this.res;
	}

	@Override
	public void paint(Graphics g1) {
		Graphics2D g = (Graphics2D) g1;
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		try {
			clearScreen(g);
			g.drawImage(offScreen1, 0, 0, this);
			for(Charge temp:Launcher.dat.getCharges()) {
				if (temp != null) {
					temp.paint(g);
				}
			}
		} catch (Throwable t) {
			Launcher.critical.log(Lev.SEVERE,
					"Unhandled exception while painting", t);
			Launcher.Exit(1);
		}
	}

	public synchronized boolean rdtDraw() throws InterruptedException,
			ConcurrentModificationException {
		if (rdt.isInterrupted())
			throw new InterruptedException("ReDraw Interrupted");

		if (o == null) {
			Launcher.info.log(Lev.WARNING,
					"Attempted paint with null offScreen graphics");
			return false;
		}
		this.res = Launcher.dat.getPeRes();
		clearScreen(o);
		if (Launcher.dat.gridOn)
			o.drawImage(getBackgroundImage(), 0, 0, this);
		if (rdt.isInterrupted())
			throw new InterruptedException("ReDraw Interrupted");
		if (Launcher.dat.fieldLinesOn) {
			drawLines(o);
			drawPC(o);
		}
		if (rdt.isInterrupted())
			throw new InterruptedException("ReDraw Interrupted");
		if (Launcher.dat.potentialLinesOn)
			drawP(o);
		if (rdt.isInterrupted())
			throw new InterruptedException("ReDraw Interrupted");
		if (Launcher.dat.ticksOn)
			drawScale(o);
		return true;
	}

	public boolean resetSize() {
		int w = this.getWidth();
		int h = this.getHeight();
		if ((w <= 0) || (h <= 0))
			return false;
		if ((offScreen == null) || (w != offScreen.getWidth(null))
				|| (h != offScreen.getHeight(null))) {
			try {
				offScreen = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
			} catch (RuntimeException e) {
				return false;
			}
		}
		if ((offScreen1 == null) || (w != offScreen1.getWidth(null))
				|| (h != offScreen1.getHeight(null))) {
			try {
				offScreen1 = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
			} catch (RuntimeException e) {
				return false;
			}
		}
		if ((Background == null) || (w != Background.getWidth(null))
				|| (h != Background.getHeight(null))) {
			try {
				Background = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
				reDrawBackground = true;
			} catch (RuntimeException e) {
				return false;
			}
		}
		o = (Graphics2D) offScreen.getGraphics();
		return true;
	}

	private void clearScreen(Graphics2D c) {
		c.setColor(Launcher.dat.backgroundColor);
		c.fillRect(0, 0, this.getWidth(), this.getHeight());
	}

	private void drawGrid(Graphics2D g) throws InterruptedException {
		g.setColor(Launcher.dat.gridColor);
		int ticSpacing = Launcher.dat.tickSpacing;
		for (int x = 0; x < this.getWidth(); x += ticSpacing) {
			if (rdt.isInterrupted())
				throw new InterruptedException("ReDraw Interrupted");
			g.drawLine(x, 0, x, this.getHeight());
		}
		for (int y = 0; y < this.getHeight(); y += ticSpacing) {
			if (rdt.isInterrupted())
				throw new InterruptedException("ReDraw Interrupted");
			g.drawLine(0, y, this.getWidth(), y);
		}
	}

	private synchronized void drawLines(Graphics2D g)
			throws InterruptedException {
		final double mult = 2 * Math.PI;
		final double n = Launcher.dat.getNumLines();
		final double dSize = (12.5);
		ThreadGroup lineThreads = new ThreadGroup("Field Line Group");
		for(Charge temp:Launcher.dat.getCharges()) {
			if (temp.getCharge() > 0) {
				double nL = n
						* (temp.getCharge() / Launcher.dat.totalPositiveCharge);
				for (int i = 0; i <= nL; i++) {
					if (restartPaint)
						return;
					double per = (i / nL);
					double x = (dSize * cos(mult * per)) + temp.getPaintX();
					double y = (dSize * sin(mult * per)) + temp.getPaintY();
					int yi = (int) (y + .5);
					int xi = (int) (x + .5);
					DrawFLineThread dtemp = new DrawFLineThread(xi, yi, g,
							lineThreads);
					dtemp.start();
				}
			}
		}
		while (lineThreads.activeCount() > 0) {
			try {
				Thread.sleep(1);
			} catch (InterruptedException e) {
			}

		}
	}

	private void drawP(Graphics2D g1) throws InterruptedException {
		int num = Launcher.dat.pLines;
		long spacing = Launcher.dat.potspacing;
		long val = 0;
		for (int r = num; r > -0; r--) {
			if (rdt.isInterrupted())
				throw new InterruptedException("ReDraw Interrupted");
			val = Launcher.dat.potOffset - (r * spacing);
			Calc.drawPELine(val, g1);
			Calc.drawPELine(-val, g1);
		}
	}

	private void drawPC(Graphics2D g) {
		for (int i = 0; i < Launcher.win.peClicks.size(); i++) {
			Calc.drawPELine((Launcher.win.peClicks.get(i)).doubleValue(), g);
		}
	}

	private void drawScale(Graphics2D g) throws InterruptedException {
		int yOff = 12;
		int xOff = 2;
		double meters = 0.0;
		g.setColor(Launcher.dat.tickColor);
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		g.setColor(Launcher.dat.text);
		for (int x = 1; x < this.getWidth(); x++) {
			if (rdt.isInterrupted())
				throw new InterruptedException("ReDraw Interrupted");
			if ((x % Launcher.dat.tickSpacing) == 0) {
				g.drawLine(x, yOff - 10, x, yOff + 5);
				if ((x % (Launcher.dat.tickSpacing * 2)) == 0) {
					meters = (x / Launcher.dat.siFactor);
					String temp = "" + (int) meters + "m";
					if (temp.length() > 4)
						temp = temp.substring(0, 4);
					g.drawString(temp, x - 7, yOff);
				}
			}
		}
		meters = 0.0;
		for (int y = 1; y < this.getWidth(); y++) {
			if (rdt.isInterrupted())
				throw new InterruptedException("ReDraw Interrupted");
			if ((y % Launcher.dat.tickSpacing) == 0) {
				g.drawLine(xOff, y, xOff + 5, y);
				if ((y % (Launcher.dat.tickSpacing * 2)) == 0) {
					meters = (y / Launcher.dat.siFactor);
					String temp = "" + (int) meters;
					if (temp.length() > 4)
						temp = temp.substring(0, 4);
					g.drawString(temp, xOff + 2, y + 3);
				}
			}
		}
	}

	private Image getBackgroundImage() {
		if (reDrawBackground)
			reDrawBackground();
		return this.Background;
	}

	private void initialize() {
		this.setBackground(Launcher.dat.backgroundColor);
		this.resetSize();
		if (offScreen != null)
			this.o = (Graphics2D) offScreen.getGraphics();
	}

	private void reDrawBackground() {
		Graphics2D g = (Graphics2D) Background.getGraphics();
		clearScreen(g);
		if (Launcher.dat.gridOn)
			try {
				drawGrid(g);
			} catch (InterruptedException e) {
				return;
			}
		reDrawBackground = false;
	}
}
